<'
# Copyright (c) 2012-2013,2017 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Thu Mar 15 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#include "autoconf/acheader.h"

#include "portLib.h"
#include "errnoLib.h"
#include "gcomLib.h"

#include "<"$comp">/c/client.h"
#include "private.h"
#include "serialize.h"
#include "typecopy.h"
#include "<"$comp">_msglib.h"

<'foreach s [$component services] {'>

/* <"[--- Service [$s name] encoding/decoding ---------------------------]"> */

int
genom_<"$comp">_<"[$s name]">_encode(
  char *buffer, int size, char *dst, int maxsize)
{
<'  if {[llength [$s parameters in inout]]} {'>
  char *p = dst;
  struct genom_<"$comp">_<"[$s name]">_input *in =
    (struct genom_<"$comp">_<"[$s name]">_input *)buffer;
  ssize_t s;
  (void)size; /* fix -Wunused-parameter */

  s = 0;
<'    foreach p [$s parameters in inout] {'>
  s += genom_serialen_<"[[$p type] mangle]">(
    <"[[$p type] pass value in->[$p name]]">);
<'    }'>
  if (s > maxsize) return ERROR;

<'    foreach p [$s parameters in inout] {'>
  genom_serialize_<"[[$p type] mangle]">(
    &p, <"[[$p type] pass value in->[$p name]]">);
<'    }'>

  return s;
<'  } else {'>
  (void)buffer; (void)size;
  (void)dst; (void)maxsize; /* fix -Wunused-parameter */
  return 0;
<'  }'>
}

int
genom_<"$comp">_<"[$s name]">_decode(
  char *buffer, int size, char *dst, int maxsize)
{
  struct genom_client_request *a = (struct genom_client_request *)dst;
<'  if {[llength [$s parameters inout out]]} {'>
  struct genom_<"$comp">_<"[$s name]">_output *out = a->output;
<'  }'>
  ssize_t len = size;
  char *p = buffer;
  int s;
  (void)maxsize; /* fix -Wunused-parameter */
  if (size == 0) return 0;

  s = 0;
  if (*p == *"") {
    p++; len--;
    a->report = genom_ok;
<'  foreach p [$s parameters inout out] {'>
    s = genom_deserialize_<"[[$p type] mangle]">(
      &p, &len, <"[[$p type] pass reference out->[$p name]]">);
<'  }'>

<'    foreach e [$s throws] {'>
  } else if (!strcmp(p, <"[$e cname]">_id)) {
    assert(a->exdetail == NULL && a->exfini == NULL);
    len -= sizeof(<"[$e cname]">_id);
    a->report = <"[$e cname]">_id;
<'      if {[llength [$e members]]} {'>
    p += sizeof(<"[$e cname]">_id);
    a->exdetail = malloc(sizeof(<"[$e declarator]">));
    if (!a->exdetail) {
      errnoSet(S_gcomLib_MALLOC_FAILED);
      return ERROR;
    }
    genom_tinit_<"[$e mangle]">(a->exdetail);
    a->exfini = (void (*)(void *))genom_tfini_<"[$e mangle]">;
    s = genom_deserialize_<"[$e mangle]">(&p, &len, a->exdetail);
<'      }'>
<'    }'>
  } else if (!strcmp(p, genom_syserr_id)) {
    assert(a->exdetail == NULL && a->exfini == NULL);
    a->exdetail = malloc(sizeof(genom_syserr_detail));
    if (!a->exdetail) {
      errnoSet(S_gcomLib_MALLOC_FAILED);
      return ERROR;
    }
    p += sizeof(genom_syserr_id);
    len -= sizeof(genom_syserr_id) + sizeof(genom_syserr_detail);
    a->report = genom_syserr_id;
    memcpy(a->exdetail, p, sizeof(genom_syserr_detail));
  } else if (!strcmp(p, genom_unkex_id)) {
    assert(a->exdetail == NULL && a->exfini == NULL);
    a->exdetail = malloc(sizeof(genom_unkex_detail));
    if (!a->exdetail) {
      errnoSet(S_gcomLib_MALLOC_FAILED);
      return ERROR;
    }
    p += sizeof(genom_unkex_id);
    len -= sizeof(genom_unkex_id) + sizeof(genom_unkex_detail);
    a->report = genom_unkex_id;
    memcpy(a->exdetail, p, sizeof(genom_unkex_detail));
  } else {
    genom_unkex_detail *d;
    assert(a->exdetail == NULL && a->exfini == NULL);
    d = a->exdetail = malloc(sizeof(genom_unkex_detail));
    if (!a->exdetail) {
      errnoSet(S_gcomLib_MALLOC_FAILED);
      return ERROR;
    }
    a->report = genom_unkex_id;
    strncpy(d->what, p, sizeof(d->what));
    d->what[sizeof(d->what)-1] = *"";
  }

  if (s) {
    a->report = genom_ok;
    switch(s) {
      case ENOMSG: errnoSet(S_gcomLib_SMALL_DATA_STR);
      case ENOMEM: errnoSet(S_gcomLib_MALLOC_FAILED);
      default: errnoSet(S_gcomLib_NOT_A_LETTER);
      }
    return ERROR;
  }

  if (len != 0) {
    a->report = genom_ok;
    errnoSet(S_gcomLib_SMALL_LETTER);
    return ERROR;
  }
  return size;
}

void
genom_<"$comp">_client_<"[$s name]">_init_input(
  struct genom_<"$comp">_<"[$s name]">_input *in)
{
<'    foreach p [$s parameters in inout] {'>
  genom_tinit_<"[[$p type] mangle]">(
    <"[[$p type] pass reference in->[$p name]]">);
<'    }'>

  /* set default value */
  genom_<"$comp">_client_<"[$s name]">_json_scan(in,
                  "<"[string map {\" \\\"} [json-param-defval $s in inout]]">",
                  NULL);
}

void
genom_<"$comp">_client_<"[$s name]">_init_output(
  struct genom_<"$comp">_<"[$s name]">_output *out)
{
<'  if {![llength [$s parameters inout out]]} {'>
  (void)out; /* fix -Wunused-parameter */
<'  }'>
<'    foreach p [$s parameters inout out] {'>
  genom_tinit_<"[[$p type] mangle]">(
    <"[[$p type] pass reference out->[$p name]]">);
<'    }'>
}

void
genom_<"$comp">_client_<"[$s name]">_fini_input(
  struct genom_<"$comp">_<"[$s name]">_input *in)
{
<'  if {![llength [$s parameters in inout]]} {'>
  (void)in; /* fix -Wunused-parameter */
<'  }'>
<'    foreach p [$s parameters in inout] {'>
  genom_tfini_<"[[$p type] mangle]">(
    <"[[$p type] pass reference in->[$p name]]">);
<'    }'>
}

void
genom_<"$comp">_client_<"[$s name]">_fini_output(
  struct genom_<"$comp">_<"[$s name]">_output *out)
{
<'  if {![llength [$s parameters inout out]]} {'>
  (void)out; /* fix -Wunused-parameter */
<'  }'>
<'    foreach p [$s parameters inout out] {'>
  genom_tfini_<"[[$p type] mangle]">(
    <"[[$p type] pass reference out->[$p name]]">);
<'    }'>
}

<'}'>
